/*
 *	HT Editor
 *	blockop.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>

#include "blockop.h"
#include "htctrl.h"
#include "htiobox.h"
#include "htkeyb.h"
#include "htstring.h"

/*
 *	CLASS ht_blockop_dialog
 */

void ht_blockop_dialog::init(bounds *b, FILEOFS pstart, FILEOFS pend, ht_list *history=0)
{
	ht_dialog::init(b, "operate on block", FS_TITLE | FS_KILLER);
	bounds c;

	bool prerange=(pend>pstart);

	ht_label *s;
	ht_statictext *q;
	
/* start */
	c=*b;
	c.h=1;
	c.w=10;
	c.x=7;
	c.y=1;
	start=new ht_strinputfield();
	start->init(&c, 64, 0);
	insert(start);
	if (prerange) {
		char t[16];
		sprintf(t, "0x%x", pstart);
		ht_inputfield_data d;
		d.isattached=0;
		d.pure.textlen=strlen(t);
		d.pure.maxtextlen=64;
		d.pure.text=(byte*)t;
		start->databuf_set(&d);
	}

/* start_desc */
	c.x=1;
	c.w=6;
	s=new ht_label();
	s->init(&c, "~start", start);
	insert(s);
	
/* end */
	c=*b;
	c.h=1;
	c.w=10;
	c.x=27;
	c.y=1;
	end=new ht_strinputfield();
	end->init(&c, 64, 0);
	insert(end);
	if (prerange) {
		char t[16];
		sprintf(t, "0x%x", pend);
		ht_inputfield_data d;
		d.isattached=0;
		d.pure.textlen=strlen(t);
		d.pure.maxtextlen=64;
		d.pure.text=(byte*)t;
		end->databuf_set(&d);
	}


/* end_desc */
	c.x=21;
	c.w=3;
	s=new ht_label();
	s->init(&c, "~end", end);
	insert(s);

/* mode */
	c=*b;
	c.h=1;
	c.w=16;
	c.x=7;
	c.y=3;
	mode=new ht_listpopup();
	mode->init(&c);
	mode->insertstring("byte (8-bit)");
	mode->insertstring("word (16-bit)");
	mode->insertstring("dword (32-bit)");
	mode->insertstring("fixed string");
//	mode->insertstring("C/ASM string");
	insert(mode);

/* mode_desc */
	c.x=1;
	c.w=12;
	c.y=3;
	s=new ht_label();
	s->init(&c, "~mode", mode);
	insert(s);
	
/* action */
	c=*b;
	c.h=1;
	c.w=30;
	c.x=7;
	c.y=6;
	action=new ht_strinputfield();
	action->init(&c, 64, 0);
	insert(action);

/* action_desc */
	c.x=1;
	c.w=27;
	c.y=5;
	s=new ht_label();
	s->init(&c, "~action: set each element to ", action);
	insert(s);
	
/* action_desc2 */
	c.x=1;
	c.w=4;
	c.y=6;
	q=new ht_statictext();
	q->init(&c, "c  =", align_left, 0);
	insert(q);
}

void ht_blockop_dialog::done()
{
	ht_dialog::done();
}

struct ht_blockop_dialog_data {
	ht_inputfield_data start;
	ht_inputfield_data end;
	ht_listpopup_data mode;
	ht_inputfield_data action;
};

/*
 *   blockop_dialog
 */

dword blockop_i;
dword blockop_c;
int blockop_vars_used;

int blockop_eval_proc(void *context, char **s, dword *v)
{
	if (**s=='i') {
		(*s)++;
		*v=blockop_i;
		blockop_vars_used++;
		return 1;
	} else if (**s=='c') {
		(*s)++;
		*v=blockop_c;
		blockop_vars_used++;
		return 1;
	}
	return 0;
}

bool progress_bit(ht_progress_indicator *progress_indicator, char *msg, FILEOFS start, FILEOFS end, FILEOFS o, FILEOFS *next_progress_draw)
{
	if ((progress_indicator) && (o>=*next_progress_draw)) {
		dword next_progress_inc=(end-start)/100;
		if (ht_keypressed()) {
			if (ht_getkey()==K_Escape) return 0;
		}

		ht_progress_indicator_data d;
		char text[64];
		d.text=text;
		if (next_progress_inc) {
			d.barpos=(o-start)/next_progress_inc;
		} else {
			d.barpos=0;
		}
		sprintf(text, msg, d.barpos);
		
		progress_indicator->databuf_set(&d);
		progress_indicator->sendmsg(msg_draw, 0);
		screen->show();
		*next_progress_draw+=next_progress_inc;
	}
	return 1;
}

bool do_fixed_str_blockop(ht_format_viewer *format, FILEOFS start, FILEOFS end, char *action, ht_progress_indicator *progress_indicator)
{
	if (end<=start) return 0;
	FILEOFS o=start;
	FILEOFS next_progress_draw=o;
	int actionlen=strlen(action);
	
/* constant expression */
	while (o<end) {
		int s=actionlen;
		if (o+s>end) s=end-o;
		format->pwrite(o, action, s);
		o+=s;
		if (!progress_bit(progress_indicator, "operating (fixed string)... %d%% complete", start, end, o, &next_progress_draw)) break;
	}
	return 1;
}

bool do_int_blockop(ht_format_viewer *format, FILEOFS start, FILEOFS end, UINT size, char *action, ht_progress_indicator *progress_indicator)
{
	if (end<=start) return 0;
	FILEOFS o=start;
	FILEOFS next_progress_draw=o;
	dword v;
	blockop_i=0;
	blockop_c=0;
	blockop_vars_used=0;
	
	if (!eval_int_expr_custom(action, &v, 0, blockop_eval_proc)) return 0;

	int tr=new_timer();
	
	start_timer(tr);

	if (blockop_vars_used) {
/* variable expression */
		while (o<end) {
			UINT s=size;
			if (o+s>end) s=end-o;
			if (format->pread(o, &blockop_c, size)!=size) return 0;
			eval_int_expr_custom(action, &v, 0, blockop_eval_proc);
			format->pwrite(o, &v, s);
			o+=s;
			blockop_i++;
			if (!progress_bit(progress_indicator, "operating (variable integer)... %d%% complete", start, end, o, &next_progress_draw)) break;
		}
	} else {
/* constant expression */
		while (o<end) {
			UINT s=size;
			if (o+s>end) s=end-o;
			if (format->pwrite(o, &v, s)!=s) return 0;
			o+=s;
			if (!progress_bit(progress_indicator, "operating (constant integer)... %d%% complete", start, end, o, &next_progress_draw)) break;
		}
	}

	stop_timer(tr);

	infobox("blockop took %d msecs.", get_timer_msec(tr));

	delete_timer(tr);
	return 1;
}

bool format_string_to_offset_if_avail(ht_format_viewer *format, byte *string, int stringlen, char *string_desc, FILEOFS *ofs)
{
	if (string && *string && stringlen<64) {
		char str[64];
		memmove(str, string, stringlen);
		str[stringlen]=0;
		if (!format->string_to_offset(str, ofs)) {
			errorbox("%s: '%s' doesn't seem to be a valid offset", string_desc, &str);
			return false;
		}
		return true;
	}
	return false;
}
		
		
void blockop_dialog(ht_format_viewer *format, FILEOFS pstart, FILEOFS pend)
{
	bounds b;
	b.w=50;
	b.h=15;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;
	
	ht_blockop_dialog *d=new ht_blockop_dialog();
	d->init(&b, pstart, pend, 0);
	if (d->run(0)) {
		ht_blockop_dialog_data t;
		d->databuf_get(&t);
		
		ht_streamfile *file=format->get_file();

		baseview->sendmsg(cmd_edit_mode_i, file, NULL);
		
		if (file->get_access_mode() & FAM_WRITE) {
			FILEOFS start=pstart, end=pend;

			if (format_string_to_offset_if_avail(format, t.start.pure.text, t.start.pure.textlen, "start", &start) &&
			format_string_to_offset_if_avail(format, t.end.pure.text, t.end.pure.textlen, "end", &end)) {
				bounds b;
				get_std_progress_indicator_metrics(&b);
				ht_progress_indicator *progress_indicator=new ht_progress_indicator();
				progress_indicator->init(&b, PIT_BAR | PIT_TEXT);

				int esize=0;
				int esizes[3]={4, 2, 1};
				switch (t.mode.cursor_id) {
/* element type: byte */
					case 0: esize++;
/* element type: word */
					case 1: esize++;
/* element type: dword */
					case 2: {
						char a[1024];	/* FIXME: possible buffer overflow */
						bin2str(a, t.action.pure.text, t.action.pure.textlen);
						esize=esizes[esize];
						if (!do_int_blockop(format, start, end, esize, a, progress_indicator)) {
							errorbox("error: '%s'", eval_int_errstr);
						}
						break;
					}
/* element type: fixed string */
					case 3: {
						char a[1024];	/* FIXME: possible buffer overflow */
						bin2str(a, t.action.pure.text, t.action.pure.textlen);
						do_fixed_str_blockop(format, start, end, a, progress_indicator);
						break;
					}
					default:
						errorbox("mode %d not supported", t.mode.cursor_id);
				}

				progress_indicator->done();
				delete progress_indicator;
			}
		}
	}

	d->done();
	delete d;
}

