/*
 *	HT Editor
 *	cstream.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "cstream.h"

#include <stdlib.h>
#include <string.h>

/*
 *	CLASS ht_cache_page
 */

ht_cache_page::ht_cache_page(FILEOFS o, dword s, dword ai)
{
	offset=o;
	size=s;
	alloc_index=ai;
	data=(byte*)malloc(size);
	if (!data) {
/* FIXME: should be generalized */
		fprintf(stderr, "in %s at line %d: out of memory !!!", __FILE__, __LINE__);
		exit(1);
	}
	read_count=0;
	write_count=0;
}

ht_cache_page::~ht_cache_page()
{
	free(data);
}

int ht_cache_page::read(FILEOFS offset, byte *buf, dword len)
{
	if (offset<size) {
		dword c=(offset+len>size) ? size-offset : len;
		memmove(buf, data+offset, c);
		read_count+=c;
		return c;
	}
	return 0;
}

int ht_cache_page::readbyte(byte *b, FILEOFS offset)
{
	if (offset<size) {
		*b=data[offset];
		read_count++;
		return 1;
	}
	return 0;
}

int ht_cache_page::writebyte(byte b, FILEOFS offset)
{
	if (offset<size) {
		data[offset]=b;
		write_count++;
		return 1;
	}
	return 0;
}

/*
 *	CLASS ht_cached_stream
 */

FILEOFS FILEOFS_delinearize(FILEOFS o)
{
//	return ((o>>16)&0xffff) | (o<<16);

	return o*0x8088405+1;	/* there's magic in here... */
	   
//	return o;
}

int compare_keys_file_delinear(ht_data *key_a, ht_data *key_b)
{
	FILEOFS a=FILEOFS_delinearize(((ht_data_uint*)key_a)->value);
	FILEOFS b=FILEOFS_delinearize(((ht_data_uint*)key_b)->value);
	return a-b;
}

/*
 *	CLASS ht_cached_file
 */

#define STD_CACHE_PAGE_SIZE 8*1024	/* MUST BE 2^n */

void ht_cached_file::init(ht_streamfile *streamfile, int own_streamfile, dword maxcsz, int fl)
{
	init(streamfile, own_streamfile, maxcsz, fl, STD_CACHE_PAGE_SIZE);
}

void ht_cached_file::init(ht_streamfile *streamfile, int own_streamfile, dword maxcsz, int fl, dword pg_gr)
{
	ht_layer_streamfile::init(streamfile, own_streamfile);
	cache_create();
	cache_mode=0;
	maxcachesize=maxcsz;
	flags=fl;
	if (flags & CF_CACHE_READS) cache_mode|=CM_READ_CACHED;
	alloc_index=0;
	offset=0;
	if (!streamfile->get_error()) size=streamfile->get_size();
	dirty_pages=0;
	page_granularity=pg_gr;
	page_mask=~(page_granularity-1);
}

void ht_cached_file::done()
{
	cache_destroy();
	ht_layer_streamfile::done();
}

void ht_cached_file::cache_create()
{
	ht_dtree *c=new ht_dtree();
	c->init(compare_keys_file_delinear, 250, 250);
	cache=c;
}

void ht_cached_file::cache_destroy()
{
	cache->destroy();
	delete cache;
}

void ht_cached_file::cache_flush()
{
	ht_data *key, *value;
	bool b=cache->enum_first(&key, &value);
	while (b) {
		ht_cache_page *c=(ht_cache_page*)value;
		cache_page_flush(c);
		b=cache->enum_next(&key, &value);
	}
}

void ht_cached_file::cache_invd()
{
	cache_destroy();
	cache_create();
}

ht_cache_page *ht_cached_file::cache_page_create(FILEOFS offset)
{
	offset&=page_mask;
	dword size=streamfile->get_size();
	if (offset>size) return 0;
	size-=offset;
	if (size>page_granularity) size=page_granularity;
	ht_cache_page *c=new ht_cache_page(offset, size, alloc_index++);
	streamfile->seek(offset);
	streamfile->read(c->data, size);
	cache->insert(new ht_data_uint(offset), c);
	return c;
}

void ht_cached_file::cache_page_destroy(ht_cache_page *c)
{
	ht_data_uint o(c->offset);
	if (c->write_count) dirty_pages--;
	cache->del(&o);
}

ht_cache_page *ht_cached_file::cache_page_find(FILEOFS offset)
{
	ht_data_uint o(offset & page_mask);
	return (ht_cache_page*)cache->get(&o);
}

void ht_cached_file::cache_page_flush(ht_cache_page *c)
{
	if ((c->write_count) && (flags & CF_CACHE_WRITES)) {
		streamfile->seek(c->offset);
		streamfile->write(c->data, c->size);
	}
	if (c->write_count) dirty_pages--;
//	c->read_count=0;
	c->write_count=0;
}

UINT ht_cached_file::cache_page_size()
{
	return sizeof (ht_cache_page)+page_granularity;
}

void ht_cached_file::discard_least_used_page()
{
#if 0
	ht_cache_page *lup=0;
	dword lup_access_count=0x7fffffff;
	dword lup_alloc_index=0x7fffffff;
	ht_data *key, *data;
	bool b=cache->enum_first(&key, &data);

	ht_cache_page *c;
	while (b) {
		c=(ht_cache_page*)data;
		if (((flags & CF_ONLY_EXPLICIT_FLUSHES) && (!c->write_count)) ||
		!(flags & CF_ONLY_EXPLICIT_FLUSHES)) {
			dword access_count=c->write_count+c->read_count;
			if (access_count<lup_access_count) {
				lup_access_count=access_count;
				lup_alloc_index=c->alloc_index;
				lup=c;
			} else if ((access_count==lup_access_count) &&
			  (c->alloc_index<lup_alloc_index)) {
				lup_alloc_index=c->alloc_index;
				lup=c;
			}
		}
		b=cache->enum_next(&key, &data);
	}
	if (lup) {
		cache_page_flush(lup);
		cache_page_destroy(lup);
	}
#else

#define MAX_FF_LUP_ACCESS_LIMIT	1000
#define MAX_FF_LUP_PAGES			5

	ht_data *key, *value;
	ht_cache_page *fp[MAX_FF_LUP_PAGES];
	UINT fp_count=0;
	bool b=cache->enum_first(&key, &value);

	ht_cache_page *c=0;
	while (b) {
		c=(ht_cache_page*)value;
		if (((flags & CF_ONLY_EXPLICIT_FLUSHES) && (!c->write_count)) ||
		!(flags & CF_ONLY_EXPLICIT_FLUSHES)) {
			dword access_count=c->write_count+c->read_count;
			if ((access_count<MAX_FF_LUP_ACCESS_LIMIT) && 
			(((flags & CF_ONLY_EXPLICIT_FLUSHES) && (!c->write_count)) ||
			!(flags & CF_ONLY_EXPLICIT_FLUSHES))) {
				fp[fp_count++]=c;
				if (fp_count>=MAX_FF_LUP_PAGES) break;
			}
		}
		b=cache->enum_next(&key, &value);
	}
	if (fp_count) {
		for (UINT i=0; i<fp_count; i++) {
			c=fp[i];
			cache_page_flush(c);
			cache_page_destroy(c);
		}
/* FIXME: dont work too well */
	} else if (c) {
		if (((flags & CF_ONLY_EXPLICIT_FLUSHES) && (!c->write_count)) ||
		!(flags & CF_ONLY_EXPLICIT_FLUSHES)) {
			cache_page_flush(c);
			cache_page_destroy(c);
		}			
	}
#endif
}

UINT ht_cached_file::get_size()
{
	return size;
}

UINT ht_cached_file::read(void *buf, UINT size)
{
	int c=0;
	if (flags & (CF_CACHE_READS | CF_CACHE_WRITES)) {
		if (size>4) {
			ht_cache_page *p;
			FILEOFS o=offset;
			while (size) {
				p=cache_page_find(o);
				if (!p) p=read_miss(o);
				dword lc=page_granularity - (o & (~page_mask));
				if (lc>size) lc=size;
				
				dword k;

				if (p) {
					k=p->read(o & (~page_mask), (byte*)buf, lc);
				} else {
					streamfile->seek(o);
					k=streamfile->read(buf, lc);
				}
				if (!k) break;
				size-=k;
				o+=k;
				c+=k;
				(byte*)buf+=k;
			}
			offset=o;
			return c;
		} else {
			for (dword i=0; i<size; i++) {
				if (!readbyte(offset+i, ((byte*)buf)+i)) break;
				c++;
			}
			offset+=c;
			return c;
		}
	}
	streamfile->seek(offset);
	c=streamfile->read(buf, size);
	offset+=c;
	return c;
}

int ht_cached_file::readbyte(FILEOFS offset, byte *b)
{
	if (flags & (CF_CACHE_READS | CF_CACHE_WRITES)) {
		ht_cache_page *c=cache_page_find(offset);
		if (!c) c=read_miss(offset);
		if (c) {
				return c->readbyte(b, offset & (~page_mask));
		} else {
			streamfile->seek(offset);
			return streamfile->read(b, 1);
		}
	}
	streamfile->seek(offset);
	return streamfile->read(b, 1);
}

ht_cache_page *ht_cached_file::read_miss(FILEOFS offset)
{
	if (flags & CF_CACHE_READS) {
		if (flags & CF_ONLY_EXPLICIT_FLUSHES) {
			if ((cache->count()-dirty_pages)*cache_page_size()>maxcachesize) discard_least_used_page();
		} else {
			if (cache->count()*cache_page_size()>maxcachesize) discard_least_used_page();
		}
		ht_cache_page *c=cache_page_create(offset);
		return c;
	}
	return 0;
}

int ht_cached_file::seek(FILEOFS o)
{
	if (o<size) {
		offset=o;
		return 1;
	}
	return 0;
}

void ht_cached_file::set_cache_mode(UINT m)
{
	if ((m & CM_READ_CACHED) && !(flags & CF_CACHE_READS)) {
/* enable read caching */
		cache_mode|=CM_READ_CACHED;
		flags|=CF_CACHE_READS;
	} else if (!(m & CM_READ_CACHED) && (flags & CF_CACHE_READS)) {
/* disable read caching */
		cache_mode&=~CM_READ_CACHED;
		flags&=~CF_CACHE_READS;
	}
}

FILEOFS ht_cached_file::tell()
{
	return offset;
}

int ht_cached_file::vcntl(UINT cmd, va_list vargs)
{
	switch (cmd) {
		case FCNTL_CACHE_INVD:
				cache_invd();
				return 0;
		case FCNTL_CACHE_FLUSH:
				cache_flush();
				return 0;
		case FCNTL_CACHE_SET_MODE:
				set_cache_mode(va_arg(vargs, UINT));
				return 0;
		case FCNTL_CACHE_GET_MODE:
				*va_arg(vargs, UINT*)=cache_mode;
				return 0;
		case FCNTL_CACHE_IS_DIRTY:
				*va_arg(vargs, bool*)=(bool)dirty_pages;
				return 0;
	   }
	   return ht_layer_streamfile::vcntl(cmd, vargs);
}

UINT ht_cached_file::write(void *buf, UINT size)
{
	int c=0;
	for (UINT i=0; i<size; i++) {
		if (!writebyte(offset+i, *(((byte*)buf)+i))) break;
		c++;
	}
	offset+=c;
	return c;
}

int ht_cached_file::writebyte(FILEOFS offset, byte b)
{
	if (flags & CF_CACHE_WRITES) {
		ht_cache_page *c=cache_page_find(offset);
		if (!c) {
			if (flags & CF_ONLY_EXPLICIT_FLUSHES) {
//          		if (read_pages*(page_granularity+sizeof (ht_cache_page))>maxcachesize) discard_least_used_page();
			} else {
				if ((cache->count())*(page_granularity+sizeof (ht_cache_page))>maxcachesize) discard_least_used_page();
			}
			c=cache_page_create(offset);
			if (!c) return 0;
			dirty_pages++;
		} else if (!c->write_count) {
			dirty_pages++;
		}
		return c->writebyte(b, offset & (~page_mask));
	}
	streamfile->seek(offset);
	return streamfile->write(&b, 1);
}

/*
 *	CLASS ht_cache_page_pbd
 */

ht_cache_page_pbd::ht_cache_page_pbd(FILEOFS offset, dword size, dword alloc_index, ht_streamfile *f)
: ht_cache_page(offset, size, alloc_index)
{
	writedirty=(byte*)malloc(size);
	file=f;
	cleanup();
}

ht_cache_page_pbd::~ht_cache_page_pbd()
{
	free(writedirty);
}

bool ht_cache_page_pbd::isdirty(FILEOFS offset, UINT range)
{
	if (range & ISDIRTY_SINGLEBIT) {
		dword m=1<<(range & ~ISDIRTY_SINGLEBIT);
		return ((writedirty[offset] & m) != (data[offset] & m));
	} else {
		for (UINT i=0; i<range; i++) {
			if (offset+i>=size) break;
			if (writedirty[offset+i]!=data[offset+i]) return 1;
		}
	}
	return 0;
}

void ht_cache_page_pbd::cleanup()
{
	FILEOFS o=file->tell();
	file->seek(offset);
	file->read(writedirty, size);
	file->seek(o);
}

/*
 *	CLASS ht_cached_file_pbd
 */

void ht_cached_file_pbd::init(ht_streamfile *streamfile, int own_streamfile, dword maxcachesize, int flags)
{
	ht_cached_file::init(streamfile, own_streamfile, maxcachesize, flags);
}

void ht_cached_file_pbd::init(ht_streamfile *streamfile, int own_streamfile, dword maxcachesize, int flags, dword page_granularity)
{
	ht_cached_file::init(streamfile, own_streamfile, maxcachesize, flags, page_granularity);
}

ht_cache_page *ht_cached_file_pbd::cache_page_create(FILEOFS offset)
{
	offset&=page_mask;
	dword size=streamfile->get_size();
	if (offset>size) return 0;
	size-=offset;
	if (size>page_granularity) size=page_granularity;
	ht_cache_page_pbd *c=new ht_cache_page_pbd(offset, size, alloc_index++, streamfile);
	streamfile->seek(offset);
	streamfile->read(c->data, size);
	cache->insert(new ht_data_uint(offset), c);
	return c;
}

void ht_cached_file_pbd::cache_page_flush(ht_cache_page *c)
{
	ht_cached_file::cache_page_flush(c);
	((ht_cache_page_pbd*)c)->cleanup();
}

UINT ht_cached_file_pbd::cache_page_size()
{
	return sizeof (ht_cache_page)+page_granularity*2;
}

bool ht_cached_file_pbd::cache_isdirty(FILEOFS offset, UINT range)
{
	if (range & ISDIRTY_SINGLEBIT) {
		ht_cache_page *c;
		c=cache_page_find(offset);
		if (c) {
			if (((ht_cache_page_pbd*)c)->isdirty(offset&(~page_mask), range)) return 1;
		}
	} else {
		ht_cache_page *c;
		int pr;
		while (range>0) {
			c=cache_page_find(offset);
			if (c) {
				pr=range;
				if ((offset&(~page_mask))+pr>c->size) pr=c->size-(offset&(~page_mask));
				if (((ht_cache_page_pbd*)c)->isdirty(offset&(~page_mask), pr)) return 1;
				offset+=pr;
				range-=pr;
			} else {
				offset+=page_granularity;
						  if (page_granularity-(offset & (~page_mask))<range) {
					range-=page_granularity-(offset & (~page_mask));
						  } else {
							range=0;
				}
				offset&=page_mask;
			}
		}
	}
	return 0;
}

int ht_cached_file_pbd::vcntl(UINT cmd, va_list vargs)
{
	switch (cmd) {
		case FCNTL_CACHE_ISDIRTY:
				FILEOFS o=va_arg(vargs, FILEOFS);
				    UINT size=va_arg(vargs, UINT);
				    bool *b=va_arg(vargs, bool*);
				    *b=cache_isdirty(o, size);
				return 0;
	   }
	   return ht_cached_file::vcntl(cmd, vargs);
}

