/*
 *	HT Editor
 *	htapp.h
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTAPP_H__
#define __HTAPP_H__

#include "htctrl.h"
#include "htdialog.h"
#include "htformat.h"

#define msg_truncate_file			HT_MESSAGE(100)

// Menu File
#define cmd_quit					HT_COMMAND(100)
#define cmd_file_load				HT_COMMAND(101)

// Menu Edit
#define cmd_clipboard_copy			HT_COMMAND(110)
#define cmd_clipboard_paste			HT_COMMAND(111)
#define cmd_clipboard_show			HT_COMMAND(112)
#define cmd_clipboard_clear			HT_COMMAND(113)
#define cmd_clipboard_copy_from_file	HT_COMMAND(114)
#define cmd_clipboard_paste_into_file	HT_COMMAND(115)

// Menu Options
#define cmd_options_palette			HT_COMMAND(120)

// Menu Windows
#define cmd_window_resizemove			HT_COMMAND(130)
#define cmd_window_close				HT_COMMAND(131)
#define cmd_window_list				HT_COMMAND(132)
#define cmd_window_log				HT_COMMAND(133)

// Menu Help
#define cmd_help_help				HT_COMMAND(140)

#define VIEWERGROUP_NAME		"viewergroup"

/*
 *	CLASS ht_status
 */
#define STATUS_DEFAULT_FORMAT "%a %L %t %d"
#define STATUS_ESCAPE '%'
#define STATUS_ANALY_ACTIVE 'a'
#define STATUS_ANALY_LINES 'L'
#define STATUS_TIME 't'
#define STATUS_DATE 'd'
#define STATUS_WORKBUFLEN 80

class ht_status: public ht_view {
protected:
	int		idle_count;
	char		*format;
	char		workbuf[STATUS_WORKBUFLEN];
	int		clear_len;
	int		analy_ani;
public:
			void init(bounds *b, char *Format);
	virtual	void done();
	virtual	void draw();
	virtual	bool idle();
private:
	virtual	char *defaultpalette();
	virtual	char *defaultpaletteclass();
			void render();
};

/*
 *	CLASS ht_keyline
 */

class ht_keyline: public ht_view {
/* overwritten */
	virtual	char *defaultpalette();
	virtual	char *defaultpaletteclass();
public:
			void init(bounds *b);
	virtual	void done();
/* overwritten */
	virtual	void draw();
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_desktop
 */

class ht_desktop: public ht_view {
protected:
/* overwritten */
	virtual	char *defaultpalette();
	virtual	char *defaultpaletteclass();
public:
			void init(bounds *b);
	virtual	void done();
/* overwritten */
	virtual	void draw();
};

/*
 *	CLASS ht_logviewer
 */

class ht_logviewer: public ht_viewer {
private:
	ht_string_list *lines;
	int maxlinecount;
	int ofs, xofs;

	void deletefirstline();
/* new */
	int cursor_up(int n);
	int cursor_down(int n);
public:
			void init(bounds *b, int maxlinecount);
	virtual	void done();
/* overwritten */
	virtual	void draw();
	virtual	void handlemsg(htmsg *msg);
/* new */
			void	insertline(char *line);
};

/*
 *	CLASS ht_app_window_entry
 */

class ht_app_window_entry: public ht_data {
public:
	ht_window *window;
	bool minimized;
	UINT number;
	UINT type;
	void *data;

	ht_app_window_entry(ht_window *window, UINT number, UINT type, bool minimized, void *data);
	~ht_app_window_entry();
};

/*
 *	CLASS ht_app
 */

class ht_app: public ht_dialog {
protected:
	ht_sorted_list *windows;

	ht_window *logwindow;
	ht_logviewer *logviewer;
	
	ht_window *helpwindow;
	
	ht_view *menu;
	ht_keyline *keyline;
	ht_desktop *desktop;

	ht_group *battlefield;
	
	int exit_program;

/* overwritten */
	virtual	char *defaultpalette();
	virtual	char *defaultpaletteclass();
public:
			void init(bounds *b);
	virtual	void done();
/* overwritten */
	virtual	void draw();
	virtual	char *func(UINT i, bool execute);
	virtual	void handlemsg(htmsg *msg);
	virtual	int load(ht_object_stream *f);
	virtual   OBJECT_ID object_id();
	virtual	int run(bool modal);
	virtual	void store(ht_object_stream *f);
/* new */
			bool create_file_window(char *filename);
			bool create_help_window(char *file, char *node);
			bool create_log_window();
			bool create_ofm_window(char *url);
			void	delete_window(ht_window *window);
			UINT find_free_window_number();
			void	focus_window(ht_window *window);
			void	insert_window(ht_window *window, UINT type, bool minimized, void *data);
			
			ht_window *get_window_by_number(UINT number);
			ht_window *get_window_by_type(UINT type);
			UINT get_window_number(ht_window *window);
			UINT get_window_listindex(ht_window *window);
			
			void log(char *line);
			void logf(char *lineformat, ...);
			void	popup_clipboard_viewer();
			void	popup_view_list();
			void	popup_view_list_dump(ht_view *view, ht_text_listbox *listbox, ht_list *structure, int depth, int *currenti, ht_view *currentv);
			void	popup_window_list();
};

/*
 *	exports
 */

#define SEARCHMODE_BIN	1
#define SEARCHMODE_CASM	2
#define SEARCHMODE_TEXT	4

void search_dialog(ht_format_viewer *format, UINT searchmodes);

extern char *globalerror;

#define LOG(a...) ((ht_app*)app)->logf(a);

/*
 *	INIT
 */

bool init_app();

/*
 *	DONE
 */

void done_app();

#endif /* __HTAPP_H__ */
