/*
 *	HT Editor
 *	htsearch.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "cstream.h"
#include "htapp.h"
#include "htctrl.h"
#include "htkeyb.h"
#include "htsearch.h"
#include "htstring.h"

#include <stdlib.h>
#include <string.h>

/*
 *	CLASS ht_fxbin_search_request
 */
 
ht_fxbin_search_request::ht_fxbin_search_request(UINT search_class, UINT flags, UINT _data_size, byte *_data) :
	ht_search_request(search_class, ST_FXBIN, flags)
{
	data_size=_data_size;
	data=(byte*)malloc(data_size);
	memmove(data, _data, data_size);
}

ht_fxbin_search_request::~ht_fxbin_search_request()
{
	free(data);
}

object *ht_fxbin_search_request::duplicate()
{
	byte *dt=(byte*)malloc(data_size);
	memmove(dt, data, data_size);
	return new ht_fxbin_search_request(search_class, flags, data_size, dt);
}

/*
 *	CLASS ht_regex_search_request
 */
 
ht_regex_search_request::ht_regex_search_request(UINT search_class, UINT flags, char *regex) :
	ht_search_request(search_class, ST_REGEX, flags)
{
	rx_str=ht_strdup(regex);
	int r=regcomp(&rx, rx_str, REG_EXTENDED | ((flags & SF_REGEX_CASEINSENSITIVE) ? REG_ICASE : 0));
	if (r) throw new ht_regex_search_exception(r, &rx);
}

ht_regex_search_request::~ht_regex_search_request()
{
	regfree(&rx);
}

object *ht_regex_search_request::duplicate()
{
	return new ht_regex_search_request(search_class, flags, rx_str);
}

/*
 *	binary search
 */

void bufdowncase(byte *buf, dword len)
{
	for (dword i=0; i<len; i++) {
		if ((buf[i]>='A') && (buf[i]<='Z')) buf[i]+=32;
	}
}

#include "htiobox.h"

#define SEARCH_BUF_SIZE 256*1024

int search_bin(ht_streamfile *file, FILEOFS start, FILEOFS end, dword pattern_len, byte *_pattern, FILEOFS *result_ofs, UINT flags, ht_progress_indicator *progress_indicator)
{
	int leave=0, success=0;

	byte *pattern=(byte*)malloc(sizeof *pattern * pattern_len);
	memmove(pattern, _pattern, sizeof *pattern * pattern_len);
	if (flags & SFBIN_CASEINSENSITIVE) bufdowncase(pattern, pattern_len);
	byte *buf=(byte*)malloc(SEARCH_BUF_SIZE);
	byte *bptr=buf;
	dword b;
	bool end_reached=0;
	FILEOFS o=start;
	FILEOFS next_progress_draw=o;
	dword next_progress_inc=(end-start)/100;

	dword pbyte=*pattern;
	file->seek(o);
	b=SEARCH_BUF_SIZE;
	if (b+o>end) b=end-o;

	dword filemode;
	   file->cntl(FCNTL_CACHE_GET_MODE, &filemode);
	   file->cntl(FCNTL_CACHE_SET_MODE, filemode & ~CM_READ_CACHED);

	int ggg2=new_timer();
	start_timer(ggg2);
	b=file->read(buf, b);
	stop_timer(ggg2);
//	LOG("debug: (search-read) %d msecs.", get_timer_msec(ggg2));
	delete_timer(ggg2);

	int ggg=new_timer();
	start_timer(ggg);
	if (flags & SFBIN_CASEINSENSITIVE) bufdowncase(buf, b);
	while (1) {
		bptr=buf;
		while ((bptr-buf+pattern_len<=b) && (bptr=(byte*)memchr(bptr, pbyte, b-pattern_len+1-(bptr-buf)))) {
			if (!memcmp(pattern, bptr, pattern_len)) {
				*result_ofs=o+bptr-buf;
				success=1;
				leave=1;
				break;
			}
			bptr++;
		}
/*		while (bptr-buf+pattern_len<=b) {
			if (pbyte==*bptr) {
				if (bufequal(pattern, bptr, pattern_len)) {
					*result_ofs=o+bptr-buf;
					success=1;
					leave=1;
					break;
				}
			}
			bptr++;
		}*/
		if (leave) break;
		if ((progress_indicator) && (o>=next_progress_draw)) {
			if (ht_keypressed()) {
				if (ht_getkey()==K_Escape) break;
			}
			
			ht_progress_indicator_data d;
			d.text=0;
			if ((end-start)/100) {
				d.barpos=(o-start)/((end-start)/100);
			} else {
				d.barpos=0;
			}
			progress_indicator->databuf_get(&d);
			progress_indicator->sendmsg(msg_draw, 0);
			screen->show();
			next_progress_draw+=next_progress_inc;
		}
		o+=b-pattern_len;
		if (end_reached) break;
		file->seek(o);
		b=SEARCH_BUF_SIZE;
		if (b+o>=end) {
			b=end-o;
			end_reached=1;
		}
		b=file->read(buf, b);
		if (flags & SFBIN_CASEINSENSITIVE) bufdowncase(buf, b);
	}
	stop_timer(ggg);

	delete pattern;
	delete buf;

//	LOG("debug: (search) %d msecs.", get_timer_msec(ggg));
	delete_timer(ggg);

	   file->cntl(FCNTL_CACHE_SET_MODE, filemode);

	return success;
}

/*
 *	CLASS ht_hexascii_search_mask
 */

void ht_hexascii_search_mask::init(bounds *b, int options, ht_list *history)
{
	ht_group::init(b, VO_SELECTABLE, "some hex/ascii search mask");
	VIEW_DEBUG_NAME("ht_hexascii_search_mask");

	bounds c;
/* ascii string */
	c.x=6;
	c.y=0;
	c.w=40;
	c.h=1;
	str=new ht_strinputfield();
	str->init(&c, 64, history);
	insert(str);
/* ascii string label */
	c.x=0;
	c.w=5;
	c.h=1;
	ht_label *strlabel=new ht_label();
	strlabel->init(&c, "~ascii", str);
	insert(strlabel);

/* hex string */
	c.x=6;
	c.y=2;
	c.w=40;
	c.h=1;
	ht_hexinputfield *hex=new ht_hexinputfield();
	hex->init(&c, 64);
	hex->attach(str);
	insert(hex);
/* hex string label */
	c.x=0;
	c.w=5;
	c.h=1;
	ht_label *hexlabel=new ht_label();
	hexlabel->init(&c, "~hex", hex);
	insert(hexlabel);

/* range start */
	c.x=10;
	c.y=4;
	c.w=10;
	c.h=1;
	range_start=new ht_strinputfield();
	range_start->init(&c, 10);
	insert(range_start);
/* range start label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *rslabel=new ht_label();
	rslabel->init(&c, "~from ofs", range_start);
	insert(rslabel);

/* range end */
	c.x=10;
	c.y=6;
	c.w=10;
	c.h=1;
	range_end=new ht_strinputfield();
	range_end->init(&c, 10);
	insert(range_end);
/* range end label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *relabel=new ht_label();
	relabel->init(&c, "~to ofs", range_end);
	insert(relabel);
/* options */
	c.w=35;
	c.y=8;
	c.h=2;
	ht_string_list *opts=new ht_string_list();
	opts->init();
	opts->insert_string("~case insensitive");
//	opts->insert_string("find all occurencies (~idle)");
	option_boxes=new ht_checkboxes();
	option_boxes->init(&c, opts);
	ht_checkboxes_data d;
	d.state=options;
	option_boxes->databuf_set(&d);
	insert(option_boxes);
}

void ht_hexascii_search_mask::done()
{
	ht_group::done();
}

/*
 *	CLASS ht_casm_search_mask
 */

void ht_casm_search_mask::init(bounds *b, int options, ht_list *history)
{
	ht_group::init(b, VO_SELECTABLE, "some c/asm search mask");
	VIEW_DEBUG_NAME("ht_casm_search_mask");

	bounds c;
/* string */
	c.x=0;
	c.y=1;
	c.w=40;
	c.h=1;
	str=new ht_strinputfield();
	str->init(&c, 64, history);
	insert(str);
/* string label */
	c.x=0;
	c.y=0;
	c.w=12;
	c.h=1;
	ht_label *strlabel=new ht_label();
	strlabel->init(&c, "C/ASM ~string", str);
	insert(strlabel);
/* hint */
	c.x=0;
	c.y=2;
	c.w=b->w-2;
	c.h=1;
	ht_statictext *hint=new ht_statictext();
	hint->init(&c, "(example: \"hello\\n\\0\\077\\xd\" 'ho',011b,66o)", 0);
	insert(hint);

/* range start */
	c.x=10;
	c.y=4;
	c.w=10;
	c.h=1;
	range_start=new ht_strinputfield();
	range_start->init(&c, 10);
	insert(range_start);
/* range start label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *rslabel=new ht_label();
	rslabel->init(&c, "~from ofs", range_start);
	insert(rslabel);

/* range end */
	c.x=10;
	c.y=6;
	c.w=10;
	c.h=1;
	range_end=new ht_strinputfield();
	range_end->init(&c, 10);
	insert(range_end);
/* range end label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *relabel=new ht_label();
	relabel->init(&c, "~to ofs", range_end);
	insert(relabel);
/* options */
	c.w=35;
	c.y=8;
	c.h=2;
	ht_string_list *opts=new ht_string_list();
	opts->init();
	opts->insert_string("~case insensitive");
//	opts->insert_string("find all occurencies (~idle)");
	option_boxes=new ht_checkboxes();
	option_boxes->init(&c, opts);
	ht_checkboxes_data d;
	d.state=options;
	option_boxes->databuf_set(&d);
	insert(option_boxes);
}

void ht_casm_search_mask::done()
{
	ht_group::done();
}

/*
 *	CLASS ht_text_search_mask
 */

void ht_text_search_mask::init(bounds *b, int options, ht_list *history)
{
	ht_group::init(b, VO_SELECTABLE, "some text search mask");
	VIEW_DEBUG_NAME("ht_text_search_mask");

	bounds c;
/* string */
	c.x=0;
	c.y=1;
	c.w=40;
	c.h=1;
	str=new ht_strinputfield();
	str->init(&c, 64, history);
	insert(str);
/* string label */
	c.x=0;
	c.y=0;
	c.w=12;
	c.h=1;
	ht_label *strlabel=new ht_label();
	strlabel->init(&c, "~regular expression", str);
	insert(strlabel);

/* range start */
	c.x=10;
	c.y=4;
	c.w=10;
	c.h=1;
	range_start=new ht_strinputfield();
	range_start->init(&c, 10);
	insert(range_start);
/* range start label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *rslabel=new ht_label();
	rslabel->init(&c, "~from addr", range_start);
	insert(rslabel);

/* range end */
	c.x=10;
	c.y=6;
	c.w=10;
	c.h=1;
	range_end=new ht_strinputfield();
	range_end->init(&c, 10);
	insert(range_end);
/* range end label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *relabel=new ht_label();
	relabel->init(&c, "~to addr", range_end);
	insert(relabel);
/* options */
	c.w=35;
	c.y=8;
	c.h=2;
	ht_string_list *opts=new ht_string_list();
	opts->init();
	opts->insert_string("~case insensitive");
//	opts->insert_string("find all occurencies (~idle)");
	option_boxes=new ht_checkboxes();
	option_boxes->init(&c, opts);
	ht_checkboxes_data d;
	d.state=options;
	option_boxes->databuf_set(&d);
	insert(option_boxes);
}

void ht_text_search_mask::done()
{
	ht_group::done();
}

